import serial
import time
import numpy as np
import math
import matplotlib.pyplot as plt


def plot():   #it plots the graph while the shutdown continues
    data = np.loadtxt(str(nomefile) + '.dat', skiprows=1, usecols=(0,1,2,3,4))
    dI = data[:,2]
    plt.clf()
    plt.errorbar((1)*(data[:,0]), (1)*(data[:,1]), dI, fmt='ro-')
    plt.title('IV Curve')
    plt.xlabel('Voltage (V)')
    plt.ylabel('Current (A)')
    plt.legend(('Idiode'),
           loc='upper center', shadow=True) 
    plt.savefig(nomefile + '.pdf')
    plt.show(block=False)
    plt.clf()
    dI = data[:,4]
    plt.errorbar((1)*(data[:,0]), (1)*(data[:,3]), dI, fmt='ro-')
    plt.title('IV Curve')
    plt.xlabel('Voltage (V)')
    plt.ylabel('Current (A)')
    plt.legend(('Iguardring'),
           loc='upper center', shadow=True) 
    plt.savefig(nomefile + '2.pdf')
    plt.show(block=False)
    plt.pause(1)
    
    

def exinterface(): #it gives back shutdown message and plots the final graph
    print('> ' + 'Shutdown.')
    print(print(70*'-'))
    plot()

def rampdown(volt,stepdown):
    if(volt >= 0):
        for z in list(reversed((range(0,math.ceil(volt)+1,int(stepdown))))):
            setvolt = 'SOUR:VOLT ' + str(z) #something is wrong here?
            gen.write(bytes(setvolt.encode('utf-8'))+b"\r\n")
            time.sleep(1)  
        voltOFF = 'SOUR:VOLT:STAT OFF'
        gen.write(bytes(voltOFF.encode('utf-8'))+b"\r\n")
        
    else:
        for z in range (math.floor(volt),1,int(stepdown)):
            setvolt = 'SOUR:VOLT ' + str(z) #something is wrong here?
            gen.write(bytes(setvolt.encode('utf-8'))+b"\r\n")
            time.sleep(1)  
        voltOFF = 'SOUR:VOLT:STAT OFF'
        gen.write(bytes(voltOFF.encode('utf-8'))+b"\r\n")

    
#Start    
print(70*'-')
print('>' + 'To stop the program and initiate abort sequence, press CTRL+C.')
print(70*'-')

gen = serial.Serial()
gen2 = serial.Serial()

# General configuration
diodeport = "COM5"
diodeport2 = "COM3"
volt     =  0  #It indicates the initial voltage 
stepvalue   = 1    #It indicates the step in the main voltage ramp 
stepdown = 1  #It indicates the step in the voltage shutdown ramp 
diodecurrlim = 2.5e-3   #It indicates the soft current limit for the diode  

try:

#Keithley communication settings
    gen.port = diodeport
    gen.stopbits = 1
    gen.timeout = 1
    gen.bytesize = 8
    gen.xonxoff = True
    gen.open()
    
#Keithley communication settings
    gen2.port = diodeport2
    gen2.stopbits = 1
    gen2.timeout = 1
    gen2.bytesize = 8
    gen2.xonxoff = True
    gen2.open()
    

except: #any exception

    print('> ' + "Problems with starting communication with the first Keithley.")
    if (gen.isOpen() or gen2.isOpen()):
        gen.close()
        gen2.close()
    exinterface()


gen.flush()
gen2.flush()


try:

    while gen.dtr != True:
        gen.readline()
    while gen2.dtr != True:
        gen2.readline()        

# define 
# start-up procedure (check instrument id, reset, other operations needed)
# based on the typical command structure del tipo voltrange = 'SOUR:VOLT:RANG 10'
# to be sent with the structure  gen.write(bytes(voltrange.encode('utf-8'))+b"\r\n")
# A response message can be read out through the structure:
# lettura = gen.readline()
# if (lettura != b''): #or (lettura2 != b''): #whenever you have an interesting response from the instrument, even if not expected
#               lettura = lettura.decode('utf-8').replace('\r\n','')

#Examples: write
    reset = '*RST' #manda reset
    gen.write(bytes(reset.encode('utf-8'))+b"\r\n")
    gen2.write(bytes(reset.encode('utf-8'))+b"\r\n")

#Examples: write and read
    identita = '*IDN?' #richiedi identita'   
    gen.write(bytes(identita.encode('utf-8'))+b"\r\n")
    lettura = gen.readline()
    gen2.write(bytes(identita.encode('utf-8'))+b"\r\n")
    lettura2 = gen2.readline()
    if (lettura != b'' and lettura != b''): #whenever you have an interesting response from the instrument, even if not expected
         lettura = lettura.decode('utf-8').replace('\r\n','')
         print('>  ' + lettura)
         lettura2 = lettura2.decode('utf-8').replace('\r\n','')
         print('>  ' + lettura2)

#Keithley operation parameters and actions
    zchOFF = 'SYST:ZCH OFF' #rimuovi zerocheck
    voltON = 'SOUR:VOLT:STAT ON' #abilita output voltage source
    voltOFF = 'SOUR:VOLT:STAT OFF' #abilita output voltage source
    voltrange = 'SOUR:VOLT:RANG 500'   #Select 10/50/500V source range
    currlimit = 'SOUR:VOLT:ILIM 2.5e-3' #Set current limit to ...
    #currrange = 'CURR:RANG 2e-9'            #Select the 200nA range
    currrange = 'CURR:RANG:AUTO ON'             #Enable auto current range
 
#Interactive input:  
    print('> ' + "Indicate the (averaged) number of measures per single voltage")
    n_misure = int(input('< '))
    while int(n_misure) < 0:
        print('***number of measures must be postive!***')
        n_misure= input('< ')
    print('> ' + "Indicate the voltage you want to reach starting from 0")
    maxvolt = input('< ')
    while float(maxvolt) > 0:
        print('***Voltage must be negative!***')
        print('> Indicate the negative voltage you want to reach (indicate the sign)')
        maxvolt = input('< ')
    print('> ' + "Indicate the step of the ramp")
    stepvalue = input('< ')
    stepvalue = float(stepvalue)
    print('> ' + "Indicate file name (.dat) where data must be saved")
    nomefile = input('<  ')
    
    #set desired configuration
    gen.write(bytes(zchOFF.encode('utf-8'))+b"\r\n")
    gen.write(bytes(voltON.encode('utf-8'))+b"\r\n")
    gen.write(bytes(voltrange.encode('utf-8'))+b"\r\n")
    gen.write(bytes(currlimit.encode('utf-8'))+b"\r\n")
    gen.write(bytes(currrange.encode('utf-8'))+b"\r\n")
    
    gen2.write(bytes(zchOFF.encode('utf-8'))+b"\r\n")
    gen2.write(bytes(voltOFF.encode('utf-8'))+b"\r\n")
    gen2.write(bytes(voltrange.encode('utf-8'))+b"\r\n")
    gen2.write(bytes(currlimit.encode('utf-8'))+b"\r\n")
    gen2.write(bytes(currrange.encode('utf-8'))+b"\r\n")

#Define file name and type
    file = open(str(nomefile) + '.dat','w')
    file.write('VoltageSource[V] I1[A] dI1[A] I2[A] dI2[A]\n')
   
#Ramp up voltage and read current at each step
    reading   = 'FORM:ELEM VSO, READ'     #Select reading, voltage source data
    gen.write(bytes(reading.encode('utf-8'))+b"\r\n")
    gen2.write(bytes(reading.encode('utf-8'))+b"\r\n")
    print('> ' + "Ramp launch")
    if(float(maxvolt) >= 0):
        miorange=range(0,int(float(maxvolt)/stepvalue)+1,1)
    else:
        miorange = list(reversed(range(int(float(maxvolt)/stepvalue),1,1)))
    for volt in miorange:
        setvolt = 'SOUR:VOLT ' + str(volt*stepvalue) 
        gen.write(bytes(setvolt.encode('utf-8'))+b"\r\n")
        time.sleep(1) #stabilize for 0.5 sec before reading values
        
        letturadati = 'READ?'
        gen.write(bytes(letturadati.encode('utf-8'))+b"\r\n") 
        gen2.write(bytes(letturadati.encode('utf-8'))+b"\r\n") 
       # time.sleep(0.2) #essential otherwise it skips a step and as a consequence provides data related to the previous voltage 
        
        letturacurr=np.zeros(n_misure)
        letturacurr2=np.zeros(n_misure)
        for i in range(0,n_misure,1):
            time.sleep(0.2)
            gen.write(bytes(letturadati.encode('utf-8'))+b"\r\n") 
            gen2.write(bytes(letturadati.encode('utf-8'))+b"\r\n") 
            lettura = gen.readline()
            lettura2 = gen2.readline()
            
            if (lettura != b'' and lettura2!=b''): #whenever you have an interesting response from the instrument, even if not expected
 
                lettura = lettura.decode('utf-8').replace('\r\n','')
                lettura2 = lettura2.decode('utf-8').replace('\r\n','')
                for k in range (0, len(lettura)):
                    if lettura[k]== ',':
                        letturacurr[i] = float(lettura[0:k])
                        break
                    # Impose current soft limit      
                if (abs(letturacurr[i])>= diodecurrlim):
                    print('\n' + '*****WARNING*****' + '\n')
                    print(letturacurr[i] + ' A \n')
                    print('*****TOO HIGH CURRENT*****' + '\n')
                    break 
                for k in range (0, len(lettura2)):
                    if lettura2[k]== ',':
                        letturacurr2[i] = float(lettura2[0:k])
                        break
                    # Impose current soft limit      
                if (abs(letturacurr2[i])>= diodecurrlim):
                    print('\n' + '*****WARNING*****' + '\n')
                    print(letturacurr2[i] + ' A \n')
                    print('*****TOO HIGH CURRENT*****' + '\n')
                    break 

        meancurr=letturacurr.mean()
        stdcurr=letturacurr.std()
        meancurr2=letturacurr2.mean()
        stdcurr2=letturacurr2.std()
        print('>  ' + 'voltage (V) = ' + str(volt*stepvalue) + ' Id (A) = ' + str(meancurr) + ' dId (A) = ' + str(stdcurr) + ' Ig (A) = ' + str(meancurr2) + ' dIg (A) = ' + str(stdcurr2))                     
        file.write(str(volt*stepvalue) + ' ' + str(meancurr) + ' ' + str(stdcurr) + ' ' + str(meancurr2) + ' ' + str(stdcurr2) + '\n')
    file.close()
    plot()
    print('> ' + "Starting shutdown procedure")

#Ramp down and close communication
    rampdown(volt*stepvalue,stepdown)
    gen.close()  
    gen2.close()
    exinterface()
 
#Define what to do at the exeption 
except: 

    print('\n' + '*****FORCED SHUTDOWN*****')
    file.close()
    plot()
    rampdown(volt*stepvalue,stepdown)
    gen.close()
    gen2.close()
    
    exinterface()