import serial
import time
import pyvisa as visa           #Connecting to the instruments using VISA
import numpy as np
import matplotlib.pyplot as plt


def exinterface(): #it gives back shutdown message and plots the final graph
    print('> Shutdown.')
    print(70*'-')
    data = np.loadtxt(str(nomefile) + '.dat')
    plt.subplot(3,1,1)
    plt.plot((-1)*(data[:,0]), data[:,1], color='green')
    plt.title('CV Curve')
    plt.xlabel('Voltage (V)')
    plt.ylabel('Capacitance (F)')
    plt.subplot(3,1,3)
    plt.plot((-1)*(data[:,0]), (1/(data[:,1])**2), color='blue')
    plt.title('(1/C**2)-V Curve')
    plt.xlabel('Voltage (V)')
    plt.ylabel('1/C**2')
    plt.show()
    

def plot():   #it plots the graph during whereas the shutdown continues
    data = np.loadtxt(str(nomefile) + '.dat')
    plt.subplot(3,1,1)
    plt.plot((-1)*(data[:,0]), data[:,1], color='green')
    plt.title('CV Curve')
    plt.xlabel('Voltage (V)')
    plt.ylabel('Capacitance (F)')
    plt.subplot(3,1,3)
    plt.plot((-1)*(data[:,0]), (1/(data[:,1])**2), color='blue')
    plt.title('(1/C**2)-V Curve')
    plt.xlabel('Voltage (V)')
    plt.ylabel('1/C**2')
    plt.show(block=False)
    plt.pause(0.001)

def rampdown() :
    for z in range (int(volt),1,int(stepdown)):
        print(str(volt) + ' ' + str(z))
        setvolt = 'SOUR:VOLT ' - str(z) 
        gen.write(bytes(setvolt.encode('utf-8'))+b"\r\n")
        time.sleep(1)  
    voltOFF = 'SOUR:VOLT:STAT OFF'
    gen.write(bytes(voltOFF.encode('utf-8'))+b"\r\n")
 
#Start    
print(70*'-')
print('>' + 'To stop the program and initiate abort sequence, press CTRL+C.')
print(70*'-')

gen = serial.Serial()
rm = visa.ResourceManager()

diodeport = 'COM5'
kerrport    = 'GPIB0::6::INSTR'
volt     = 0   #It indicates the initial voltage 
stepvalue   = 0.5   #It indicates the step in the voltage increasing ramp 
stepdown = 10   #It indicates the step in the voltage decreasing ramp 
diodecurrlim = 0.000002   #It indicates the arbitrary current limit for the diode  


try:
    kerr = rm.open_resource(kerrport)

except:  #wrong port
    print('>' + "Problems with starting communication with the Kerr.")
    kerr.close()
    exinterface() 


try:

    gen.port = diodeport
    gen.stopbits = 1
    gen.timeout = 1
    gen.bytesize = 8
    gen.xonxoff = True
    gen.open()
    

except: #any exception

    print('>' + "Problems with starting communication with the Keithley.")
    if gen.isOpen():
        gen.close()
    exinterface()

gen.flush()

kerr.write('*RST') 
kerr.write('*ESE 1;*SRE 32;*CLS;:FORM ASC') # *ESE 1 - enables OPC reporting # *SRE 32 -enables standard event bits for OPC reporting

# Set calibration
kerr.write(":CAL")            #Select calibrate mode / path.
print('> Do you want to perform self calibration? y/n')
x = input('<')
if x == 'y':
    print('>' + 'Are all BNCs disconnected? y/n')
    x = input('<')
    while x != 'y':
        print('>' + 'Disconnect all BNCs')
        print('>' + 'Are all BNCs disconnected now? y/n')
        x= input('<')
    kerr.write(":CAL:SELF-CAL")   #Perform self-calibration; disconnect all BNCs from the instrument terminals before using this command.
    print('>' + 'Wait for the end of the self calibration')

print('> Do you want to perform open circuit trimming? y/n')
x = input('<')
if x == 'y':
    print('> Connect all BNCs and open the circuit')
    print('>' + 'Are all BNCs connected? y/n')
    x = input('<')
    while x != 'y':
        print('>' + 'Are you ready now? y/n')
        x= input('<')
    kerr.write(":CAL:OC-TRIM 4")  #Perform open circuit trimming.
    print('>' + 'Wait for the end of the O/C trimming then close the circuit')
    
print('>' + 'Is circuit closed and everything ready for the measurement? y/n')
x=input('<')
while x != 'y':
    print('>' + 'Connect all BNCs')
    print('>' + 'Are all BNCs connected now? y/n')
    x=input('<')


try:

    #it starts a cycle from which you only exit when dtr == True which means:
    #    (1) that input buffer is full, improbable because you immediately read
    #        the output and  and there should be no pending commands in the buffer
    #    (2) che lo strumento sta cercando di inviarmi una risposta a una query,
    #        ma anche questo è difficile se io leggo tutto per ordine
    while gen.dtr != True:
        gen.readline() #(1) e (2)

    identita = '*IDN?'
    reset = '*RST' 
    zchOFF = 'SYST:ZCH OFF'
    voltON = 'SOUR:VOLT:STAT ON'
    
    gen.write(bytes(reset.encode('utf-8'))+b"\r\n")
    gen.write(bytes(zchOFF.encode('utf-8'))+b"\r\n")
    gen.write(bytes(voltON.encode('utf-8'))+b"\r\n")
    gen.write(bytes(identita.encode('utf-8'))+b"\r\n")
    
        #gen.read(45) up to 45 characters (of one byte each)
        #if you prefer to read a line of indefinite length, then: 
    lettura = gen.readline()

    if (lettura != b''): #whenever you have an interesting response from the instrument, even if not expected
         lettura = lettura.decode('utf-8').replace('\r\n','')
         print('>' + lettura)
         print('>' + kerr.query('*IDN?'))

    #Select the required operating mode
    kerr.write(":MEAS")          #Go to measurement mode.
    kerr.write(":MEAS:TEST:AC")  #Select AC measurements.
    kerr.write(":TERM 2")        #Select 2 terminal measurements.
 
    #Setting Kerr parameters
    kerr.write(":MEAS:FREQ 100k")  #Set frequency of AC measurement. 
    kerr.write(":MEAS:LEV 0.5V")     #Set drive level for currently selected test.
    kerr.write(":MEAS:EQU-CCT PAR")  #Select the equivalent circuit type for AC tests.
    #kerr.write(":MEAS:LEV AUTO")  #Select the required measurement range condition for AC tests.
    kerr.write(":MEAS:NOMINAL 1e-12F")   
    ########## :MEAS:SPEED <disc> devo settarlo? #Select the required measurement speed. 
    ########## :MEAS:ALC <disc> devo settarlo?   #Select the state of Automatic Level Control for AC tests. 


    #Start querying
    #alc   = kerr.query(":MEAS:ALC?")   #Query the ALC setting.
    #range = kerr.query(":MEAS:RANGE?")   #Query the current measurement range.
    #speed = kerr.query(":MEAS:SPEED?")   #Query the current test speed.
    freq  = kerr.query(":MEAS:FREQ?")  #Query the AC frequency.
    level = kerr.query(":MEAS:LEV?")   #Query the AC level.

    #Print the status of the major settings.
    print('>' + "AC Frequency = " + freq[0:-1] + " Hz")   #Print the AC frequency.
    print('>' + "AC Drive level = " + level[0:-1] + " V") #Print the AC level.


    #Settings bias voltage
    kerr.query(":MEAS:BIAS VEXT")  #Turn on voltage bias.
    kerr.query(":MEAS::BIAS OFF")   #Select external voltage bias drive.
    
    
    #Setting Keithley parameters
    voltrange = 'SOUR:VOLT:RANG 500'   #Select 500V source range
    gen.write(bytes(voltrange.encode('utf-8'))+b"\r\n")
    currlimit = 'SOUR:VOLT:ILIM 250e-6' #Set current limit to 250uA
    gen.write(bytes(currlimit.encode('utf-8'))+b"\r\n")
    
    #currrange = 'RANG 200e-9'            #Select the 200nA range
    autorang = 'RANG:AUTO ON'             #Enable auto current range
    gen.write(bytes(autorang.encode('utf-8'))+b"\r\n")
    

    print('> ' + "Indicate the positive voltage you want to provide to the n side (indicate the sign and check connections)")
    maxvolt = input('<')
    while int(maxvolt) < 0:
        print('***Voltage must be positive: biasing n side with positive voltage! Check your connections ***')
        print('> ' + 'Indicate the positive voltage you want to provide to the n side (indicate the sign and check connections)')
        maxvolt = input('< ')
        
    print('> ' + "Indicate file name (.dat) where data must be saved")
    nomefile = input('<')

    file = open(str(nomefile) + '.dat','w')
    reading   = 'FORM:ELEM  VSO, READ'     #Select reading, voltage source data
    gen.write(bytes(reading.encode('utf-8'))+b"\r\n") 
    kerr.write(':MEAS:FUNC:C')

    print('> ' + "Ramp launch")
    
   
    for volt in range (0,int(int(maxvolt)/stepvalue)+1,1):
        #print(maxvolt, stepvalue, int(maxvolt)/stepvalue)
        setvolt = 'SOUR:VOLT ' + str(volt*stepvalue) 
        gen.write(bytes(setvolt.encode('utf-8'))+b"\r\n")
        time.sleep(1)
        
        letturadati = 'READ?'
        
        
        gen.write(bytes(letturadati.encode('utf-8'))+b"\r\n") 
        
        time.sleep(1) #essential otherwise it skips a step and as a consequence provides data relating to the previous voltage
        lettura  = gen.readline()
        capacita = kerr.query(':MEAS:TRIG') 
        #print(kerr.query(':MESS?'))
      #  if (kerr.query(':MESS?') and 0x4000):
       #     print("communication error")
       #     time.sleep(1)
       #     capacita = kerr.query(':MEAS:TRIG') 
       #if get error repeat query
        if (lettura != b''): #whenever you have an interesting response from the instrument, even if not expected
                lettura = lettura.decode('utf-8').replace('\r\n','')
                for k in range (0, len(lettura)):
                    if lettura[k]== ',':
                      letturacurr = lettura[0:k]
                      letturavolt = lettura[k+1:] 
                      break
                for n in range (0, len(capacita)):
                    if capacita[n]== ',':
                      letturacap = capacita[0:n]
                      break
                #letturavolt = float(letturavolt)*50 # needed only on the faulty keithley - keiR
                #letturavolt = str(letturavolt)# needed only on the faulty keithley - keiR
                print('> voltage (V) = ' + str(volt*stepvalue) + ' capacitance (F) = ' + letturacap + ' current (A) = ' + letturacurr)
                file.write(str(volt*stepvalue) + ' ' + letturacap + ' ' + letturacurr + '\n')
                if (abs(float(letturacurr))>= float(diodecurrlim)):
                    print('\n' + '*****WARNING*****' + '\n')
                    print(letturacurr + ' A \n')
                    print('*****TOO HIGH CURRENT*****' + '\n')
                    break
    file.close()
    plot()
    print('> ' + "Starting shutdown procedure")
    rampdown() 
    gen.close()
    kerr.close()
  
    exinterface()
    
except: 

    print('\n' + '*****FORCED SHUTDOWN*****')
    file.close()
    plot()
    rampdown()
    gen.close()
    kerr.close()
    
    exinterface()
